from typing import Tuple, List, Union
from numpy import ndarray

NoneType = type(None)
Point = Tuple[float, float]  # (x, y)
Contour = ndarray
Circle = Tuple[Tuple[float, float], float]  # ((center_x, center_y), radius)
Rect = Tuple[float, float, float, float]  # (x, y, width, height)
Polygon = Contour
FixedPolygon = List[Point]
RotatedRect = Tuple[Point, Point, float]  # ((x1, y1), (x2, y2), angle)
Ellipse = RotatedRect
Frame = Union[ndarray, NoneType]
Color = Tuple[int, int, int]



import numpy as np

from gbvision.constants.system import EMPTY_PIPELINE
from gbvision.models.contours import find_contours, FilterContours, sort_contours, contours_to_rotated_rects_sorted
from gbvision.models.shapes import filter_inner_rotated_rects
from .object_finder import ObjectFinder


class RotatedRectFinder(ObjectFinder):
    """
    finds a rectangular object, but rotated. recommended to use when you know the shape isn't parallel to the camera
    """

    def __init__(self, threshold_func, game_object, area_scalar=1.0, contour_min_area=3.0):
        """

        :param area_scalar: optional, a scalar to multiply the area by, for fine tuning of the function's output
        :param contour_min_area: the minimal area of a contour, used for FilterContours, default is 0 (no area limit)
        """
        ObjectFinder.__init__(self, threshold_func, game_object)
        self._full_pipeline = (EMPTY_PIPELINE +
                               threshold_func +
                               find_contours +
                               FilterContours(min_area=contour_min_area) +
                               sort_contours +
                               contours_to_rotated_rects_sorted +
                               filter_inner_rotated_rects)
        self.area_scalar = area_scalar

    def __call__(self, frame, camera):
        rects = self._full_pipeline(frame)
        return list(map(
            lambda rect: self.game_object.location_by_params(camera,
                                                             self.area_scalar * np.sqrt(rect[1][0] * rect[1][1]),
                                                             rect[0]), rects))


import abc
from threading import Thread
from typing import Tuple

import numpy as np

from gbvision.constants.types import Frame
from .camera import Camera


class AsyncCamera(Camera, abc.ABC):
    """
    an abstract class that represents an async camera
    the async camera is similar to a regular camera, but executes the read actions on another thread
    thus not blocking the processing thread
    """

    def __init__(self):
        self.__ok, self.__frame = False, None
        self.__thread = Thread(target=self.__async_read_wrapper)
        self.__thread.start()

    def read(self, image=None):
        return self.__ok, self.__frame

    @abc.abstractmethod
    def _read(self) -> Tuple[bool, Frame]:
        """
        reads from the camera synchronously (similar to Camera.read), unsafe, not to use by the programmer
        
        :return: tuple of bool (indicates if read was successful) and the frame (if successful, else None)
        """
        pass

    def __async_read_wrapper(self):
        while self.is_opened():
            self.__ok, self.__frame = self._read()



from .recording_wrapper_opencv_window import RecordingWrapperOpenCVWindow
from .camera_window import CameraWindow


class RecordingCameraWindow(RecordingWrapperOpenCVWindow, CameraWindow):
    """
    a basic window that displays the stream from a stream receiver
    """




from .wrapper_opencv_window import WrapperOpenCVWindow


class CameraWindow(WrapperOpenCVWindow):
    """
    a basic window that displays a feed from a camera
    in this class, self.wrap_object will be of type Camera
    
    :type self.wrap_object: gbvision.Camera
    """
    def _get_frame(self):
        _, frame = self.wrap_object.read()
        return frame
